
from .rdp_accountant import compute_rdp, get_privacy_spent,compute_rdp_single,find_sigma_for_target_epsilon

class DP_controller:
    def __init__(self,data_size, batch_size, total_steps, epsilon_budget,delta=None,clip_coeff = 1, sigma = None,device = 'cpu',verbose = True):
        self.data_size = data_size
        self.batch_size = batch_size
        self.total_steps = total_steps # might step the discriminator more times in one epoch

        self.epsilon_budget = epsilon_budget
        self.delta = delta
        self.clip_coeff = clip_coeff
        self.sigma = sigma
        self.device = device
        self.verbose = verbose


        if not isinstance(batch_size, int) or batch_size < 1: 
            raise ValueError("Batch size must be a positive integer")

        if not isinstance(total_steps, int) or total_steps < 1:
            raise ValueError("Total steps must be a positive integer")
        
        self.q = batch_size/data_size if batch_size < data_size else 1 # If the batch size is set greater and the toal data size, then the whole data is used


        max_lmbd = 4095
        self.lmbds = range(2, max_lmbd + 1)

        if clip_coeff is None: raise ValueError("Clip coefficient cannot be set to None")

        self.delta = self._get_optimal_delta(self.data_size)

        self.sigma = self._get_optimal_sigma(self.q, self.total_steps,self.epsilon_budget, self.delta, self.sigma)

        self.rdp = compute_rdp_single(self.q, self.sigma, self.lmbds)

    

    def _get_optimal_delta(self,data_size):
        if self.delta:
            if self.verbose:
                print(f"Using provided delta: {self.delta}")
            return self.delta
        else:
            if self.verbose:
                print(f"No delta provided, setting delta: {0.1/data_size}")
            return 0.1/data_size


    def _get_optimal_sigma(self,q,total_steps, epsilon_budget, delta, sigma):
        if delta is None: ValueError("Delta must be provided")
        if sigma:
            if self.verbose:
                print(f"Using provided sigma: {sigma}, potentially overwriting epsilon budget")
            return sigma
        
        if epsilon_budget is None: raise ValueError("Either epsilon budget or must be provided")
        sigma = find_sigma_for_target_epsilon(q, total_steps,epsilon_budget, delta,verbose=self.verbose)
        if self.verbose:
            print(f"Sigma set to: {sigma}, using epsilon budget: {epsilon_budget} and delta: {delta}")
        return sigma

    def _compute_single_rdp(self,q, sigma, total_steps, lmbds):
        if total_steps <= 0: raise ValueError("Total steps must be a positive integer")
        if lmbds is None: raise ValueError("Lambdas must be provided")
        return compute_rdp_single(q, sigma, total_steps, lmbds)

    def get_spent_privacy(self,steps):
        total_rdp = self.rdp * steps
        privacy_spent, delta, _ = get_privacy_spent(self.lmbds, total_rdp, target_delta=self.delta)
        return privacy_spent

    def get_sigma(self):
        return self.sigma

    def get_clip_coeff(self):
        return self.clip_coeff

        
        



        
        


import numpy as np
import pandas as pd
from sklearn import preprocessing
from sklearn import model_selection

class DataPrep(object):
  
    def __init__(self, raw_df: pd.DataFrame, categorical: list, log:list, mixed:dict, general:list, non_categorical:list, integer:list, type:dict, test_ratio:float):
        
        
        self.categorical_columns = categorical
        self.log_columns = log
        self.mixed_columns = mixed
        self.general_columns = general
        self.non_categorical_columns = non_categorical
        self.integer_columns = integer
        self.column_types = dict()
        self.column_types["categorical"] = []
        self.column_types["mixed"] = {}
        self.column_types["general"] = []
        self.column_types["non_categorical"] = []
        self.lower_bounds = {}
        self.label_encoder_list = []
        
        problem = list(type.keys())[0]
        target_col = list(type.values())[0]
        if problem:
            y_real = raw_df[target_col]
            X_real = raw_df.drop(columns=[target_col])
            
            if problem=="Classification":
                X_train_real, _, y_train_real, _ = model_selection.train_test_split(X_real ,y_real, test_size=test_ratio, stratify=y_real,random_state=42)
            else:
                X_train_real, _, y_train_real, _ = model_selection.train_test_split(X_real ,y_real, test_size=test_ratio,random_state=42) 
            
            X_train_real[target_col]= y_train_real

            self.df = X_train_real

        else:
            self.df = raw_df
        self.df = self.df.replace(r' ', np.nan)
        self.df = self.df.fillna('empty')
       
        all_columns= set(self.df.columns)
        irrelevant_missing_columns = set(self.categorical_columns)
        relevant_missing_columns = list(all_columns - irrelevant_missing_columns)
        
        for i in relevant_missing_columns:
            if i in self.log_columns:
                if "empty" in list(self.df[i].values):
                    self.df[i] = self.df[i].apply(lambda x: -9999999 if x=="empty" else x)
                    self.mixed_columns[i] = [-9999999]
            elif i in list(self.mixed_columns.keys()):
                if "empty" in list(self.df[i].values):
                    self.df[i] = self.df[i].apply(lambda x: -9999999 if x=="empty" else x )
                    self.mixed_columns[i].append(-9999999)
            else:
                if "empty" in list(self.df[i].values):   
                    self.df[i] = self.df[i].apply(lambda x: -9999999 if x=="empty" else x)
                    self.mixed_columns[i] = [-9999999]
        
        if self.log_columns:
            for log_column in self.log_columns:
                valid_indices = []
                for idx,val in enumerate(self.df[log_column].values):
                    if val!=-9999999:
                        valid_indices.append(idx)
                eps = 1
                lower = np.min(self.df[log_column].iloc[valid_indices].values)
                self.lower_bounds[log_column] = lower
                if lower>0: 
                    self.df[log_column] = self.df[log_column].apply(lambda x: np.log(x) if x!=-9999999 else -9999999)
                elif lower == 0:
                    self.df[log_column] = self.df[log_column].apply(lambda x: np.log(x+eps) if x!=-9999999 else -9999999) 
                else:
                    self.df[log_column] = self.df[log_column].apply(lambda x: np.log(x-lower+eps) if x!=-9999999 else -9999999)

        for column_index, column in enumerate(self.df.columns):            
            if column in self.categorical_columns:        
                label_encoder = preprocessing.LabelEncoder()
                self.df[column] = self.df[column].astype(str)
                label_encoder.fit(self.df[column])
                current_label_encoder = dict()
                current_label_encoder['column'] = column
                current_label_encoder['label_encoder'] = label_encoder
                transformed_column = label_encoder.transform(self.df[column])
                self.df[column] = transformed_column
                self.label_encoder_list.append(current_label_encoder)
                self.column_types["categorical"].append(column_index)

                if column in self.general_columns:
                    self.column_types["general"].append(column_index)
            
                if column in self.non_categorical_columns:
                    self.column_types["non_categorical"].append(column_index)
            
            elif column in self.mixed_columns:
                self.column_types["mixed"][column_index] = self.mixed_columns[column]
            
            elif column in self.general_columns:
                self.column_types["general"].append(column_index)
            

        super().__init__()
        
    def inverse_prep(self, data, eps=1):
        
        df_sample = pd.DataFrame(data,columns=self.df.columns)
     
        for i in range(len(self.label_encoder_list)):
            le = self.label_encoder_list[i]["label_encoder"]
            df_sample[self.label_encoder_list[i]["column"]] = df_sample[self.label_encoder_list[i]["column"]].astype(int)
            df_sample[self.label_encoder_list[i]["column"]] = le.inverse_transform(df_sample[self.label_encoder_list[i]["column"]])

        if self.log_columns:
            for i in df_sample:
                if i in self.log_columns:
                    lower_bound = self.lower_bounds[i]
                    if lower_bound>0:
                        df_sample[i].apply(lambda x: np.exp(x)) 
                    elif lower_bound==0:
                        df_sample[i] = df_sample[i].apply(lambda x: np.ceil(np.exp(x)-eps) if (np.exp(x)-eps) < 0 else (np.exp(x)-eps))
                    else: 
                        df_sample[i] = df_sample[i].apply(lambda x: np.exp(x)-eps+lower_bound)

        if self.integer_columns:
            for column in self.integer_columns:
                df_sample[column]= (np.round(df_sample[column].values))
                df_sample[column] = df_sample[column].astype(int)

        df_sample.replace(-9999999, np.nan,inplace=True)
        df_sample.replace('empty', np.nan,inplace=True)

        return df_sample



"""
Generative model training algorithm based on the CTABGANSynthesiser

"""
import pandas as pd
import time
from model.pipeline.data_preparation import DataPrep
from model.synthesizer.ctabgan_synthesizer import CTABGANSynthesizer

from model.pipeline2.data_type_assigner import Data_type_assigner
from model.pipeline2.data_preparation import DataPrep as DataPrep2
from model.pipeline2.null_value_transformer import Null_value_transformer


import warnings

warnings.filterwarnings("ignore")

class CTAB_XTRA_DP():

    def __init__(self,
                 df,
                 test_ratio = 0.20,
                 categorical_columns = [], 
                 log_columns = [],
                 mixed_columns= {},
                 general_columns = [],
                 non_categorical_columns = [],
                 integer_columns = [],
                 problem_type= {},
                 dp_constraints = {}
                 ):

        self.__name__ = 'CTABGAN'
              
        self.synthesizer = CTABGANSynthesizer()
        self.raw_df = df
        self.test_ratio = test_ratio
        self.dp_constraints = dp_constraints
        self.categorical_columns = categorical_columns
        self.log_columns = log_columns
        self.mixed_columns = mixed_columns
        self.general_columns = general_columns
        self.non_categorical_columns = non_categorical_columns
        self.integer_columns = integer_columns
        self.problem_type = problem_type
                
    def fit(self,epochs= 100):
        
        start_time = time.time()
        
        self.data_type_assigner = Data_type_assigner(self.raw_df, self.integer_columns,self.mixed_columns)
        
        self.null_value_transformer = Null_value_transformer()
        self.mixed_columns = self.null_value_transformer.fit(self.raw_df, self.categorical_columns, self.mixed_columns)
        self.raw_df = self.null_value_transformer.transform(self.raw_df)

        self.data_prep2 = DataPrep2(self.raw_df,self.categorical_columns,self.log_columns)
        self.prepared_data = self.data_prep2.preprocesses_transform(self.raw_df)

        self.synthesizer.fit(train_data=self.prepared_data,
                    dp_constraints=self.dp_constraints, 
                     categorical=self.categorical_columns, 
                     mixed=self.mixed_columns, 
                     general=self.general_columns, 
                     non_categorical=self.non_categorical_columns, 
                     type=self.problem_type, 
                     epochs=epochs)
        end_time = time.time()
        print('Finished training in',end_time-start_time," seconds.")


    def generate_samples(self,n=1000):
        
        sample_transformed = self.synthesizer.sample(n)#self.synthesizer.sample(n, column_index, column_value_index)
        sample_transformed = pd.DataFrame(sample_transformed, columns=self.prepared_data.columns)
        
        sample = self.data_prep2.preprocesses_inverse_transform(sample_transformed)
        sample = self.null_value_transformer.inverse_transform(sample)
        sample_with_data_types = self.data_type_assigner.assign(sample)
        
        return sample_with_data_types



import numpy as np
import torch
from .column_transformer import Column_transformer

class General_transformer(Column_transformer):
    def __init__(self, column):
        super().__init__()
        self.min = column.min()
        self.max = column.max()

        self.fit(column.to_numpy())

    def fit(self, data_col):
        self.model = None 
        self.components = None
        self.output_info = [(1, 'tanh','yes_g')]
        self.output_dim = 1

    def transform(self, data_col):
        
        current = (data_col - (self.min)) / (self.max  - self.min) * 2 - 1
        current = current.reshape([-1, 1])
        return current

    def inverse_transform(self, data,st):
        u = data[:, st]
        u = (u + 1) / 2
        u = np.clip(u, 0, 1)
        u = u * (self.max - self.min) + self.min
        new_st = st + 1
        return u, new_st, []

    def inverse_transform_static(self, data, transformer, st, device, n_clusters=10): #TODO: remove n cluster when finding out how to handle it
        u = data[:, st]
        u = (u + 1) / 2
        u = torch.clamp(u, 0, 1)
        u = u * (transformer.max - transformer.min) + transformer.min
        new_st = st + 1
        return u, new_st, []
        

        
        




import numpy as np
import pandas as pd
import torch
from typing import List
from ..pipeline.data_preparation import DataPrep

from .column_transformers.column_transformer import Column_transformer
from .column_transformers.gmm_transformer import GMM_transformer
from .column_transformers.categorical_transformer import Categorical_transformer
from .column_transformers.mixed_data_transformer import Mixed_data_transformer
from .column_transformers.gaussian_transformer import Gaussian_transformer
from .column_transformers.general_transformer import General_transformer
from .column_transformers.gaussian_truncated_transformer import Gaussian_truncated_transformer


class DataTransformer():
    
    def __init__(self, prepared_data, categorical_list, mixed_dict, general_list,truncated_gaussian_columns, n_clusters=10, eps=0.005):
        self.meta = None
        self.n_clusters = n_clusters
        self.eps = eps
        
        #self.data_prep = data_prep
       
        self.transformers = self.setup_transformers(prepared_data,categorical_list, mixed_dict, general_list,truncated_gaussian_columns)
        

    

    
    def get_column_transform(self, data,column, categorical_list, mixed_dict, general_list,truncated_gaussian_columns):
        if column in categorical_list: return Categorical_transformer(data[column])
        if column in mixed_dict: return Mixed_data_transformer(data[column],mixed_dict[column])
        if column in general_list: return General_transformer(data[column])
        if column in truncated_gaussian_columns: return Gaussian_truncated_transformer(data[column])
        return GMM_transformer(data[column])  # No type specified by user


    def setup_transformers(self,data,categorical_list, mixed_dict, general_list,truncated_gaussian_columns = []):
        transformers = []
        for column in data.columns:
            column_type = self.get_column_transform(data,column, categorical_list, mixed_dict, general_list,truncated_gaussian_columns)
            transformers.append(column_type)
        return transformers


            
    def fit_transformers(self,data):
        assert self.transformers is not None, "Transformers are not initialized"
        assert len(self.transformers) == data.shape[1], "Mismatch between number of transformers and data columns"
        for i in range(len(self.transformers)):
            transformer = self.transformers[i]
            transformer.fit(data.iloc[:,i].to_numpy())
        return True


       
    def transform(self, data):
        transforms = []
        for i, transformer in enumerate(self.transformers):
            if transformer is None: continue
            transforms.append(transformer.transform(data.iloc[:,i].to_numpy()))

        return np.concatenate(transforms, axis=1)

    def inverse_transform(self, data):
        data_t = np.zeros([len(data), len(self.transformers)])
        all_invalid_ids = []
        st = 0
        for idx, transformer in enumerate(self.transformers):
            new_data, st, invalid_ids = transformer.inverse_transform(data, st)
            data_t[:,idx] = new_data
            all_invalid_ids += invalid_ids
            print("transformer", idx    , "invalid ids", invalid_ids)
        all_invalid_ids = np.unique(all_invalid_ids)
        data_t = np.delete(data_t, list(all_invalid_ids), axis=0)
        return data_t, len(all_invalid_ids)


    def inverse_transform_static(self,data, transformer, device,n_clusters=10):
        gaussian_columns = []
        transformers = transformer.get_transformers()
        data_t = torch.zeros(len(data), len(transformers), device=device)
        all_invalid_ids = []
        st = 0
        for id_, transformer in enumerate(transformers):
            new_data, st, invalid_ids = transformer.inverse_transform_static(data, transformer, st,device,n_clusters)
            data_t[:,id_] = new_data
            all_invalid_ids += invalid_ids

        mask = torch.ones(len(data_t), dtype=torch.bool, device=device)
        mask[list(all_invalid_ids)] = False


        data_t = data_t[mask]

        return data_t, len(all_invalid_ids)

    def get_transformers(self): # TODO: should be removed if possible with the inverse backporpagation
        return self.transformers
         
    def get_output_info(self):
        return [transformer.get_output_info() for transformer in self.transformers]

    def get_output_dim(self):
        return sum([transformer.get_output_dim() for transformer in self.transformers])

    def get_components(self): #TODO: check if needed
        return [transformer.get_components() for transformer in self.transformers]

    def get_output_info_flat(self):
        output_info_list = []
        for transformer in self.transformers:
            output_info = transformer.get_output_info()
            if len(output_info) == 1:
                output_info_list.append(output_info[0])
            else:
                output_info_list.extend(output_info)
        return output_info_list


   
            

    

class ImageTransformer():

    def __init__(self, side):
    
        self.height = side
            
    def transform(self, data):

        if self.height * self.height > len(data[0]):
            
            padding = torch.zeros((len(data), self.height * self.height - len(data[0]))).to(data.device)
            data = torch.cat([data, padding], axis=1)

        return data.view(-1, 1, self.height, self.height)

    def inverse_transform(self, data):
        
        data = data.view(-1, self.height * self.height)

        return data





import numpy as np
import pandas as pd
import torch
from sklearn.mixture import BayesianGaussianMixture
from .column_transformer import Column_transformer
class GMM_transformer(Column_transformer):
    def __init__(self, column,n_clusters=10,eps=0.005):
        super().__init__()
        self.n_clusters = n_clusters
        self.eps = eps
        self.min = column.min()
        self.max = column.max()
        
        self.fit(column.to_numpy())


    def fit(self, data_col):
        gm = BayesianGaussianMixture(
                n_components = self.n_clusters, 
                weight_concentration_prior_type='dirichlet_process',
                weight_concentration_prior=0.001, 
                max_iter=100,n_init=1, random_state=42)
        gm.fit(data_col.reshape([-1, 1]))
        mode_freq = (pd.Series(gm.predict(data_col.reshape([-1, 1]))).value_counts().keys())
        
        old_comp = gm.weights_ > self.eps

        # The final components are the ones that are frequent and have a weigh greater than the eps cutoff
        self.components = [(i in mode_freq) & old_comp[i] for i in range(self.n_clusters)]
        self.output_info = [(1, 'tanh','no_g'), (np.sum(self.components), 'softmax',None)]
        self.output_dim = 1 + np.sum(self.components)
        self.model = gm
        

    def transform(self, data_col):
        np.random.seed(22) # TODO: remove used for debugging
        data_col = data_col.reshape([-1, 1])
        means = self.model.means_.reshape((1, self.n_clusters))
        stds = np.sqrt(self.model.covariances_).reshape((1, self.n_clusters))
        features = np.empty(shape=(len(data_col),self.n_clusters))
       
        features = (data_col - means) / (4 * stds)

        probs = self.model.predict_proba(data_col.reshape([-1, 1]))
        n_opts = sum(self.components)
        features = features[:, self.components]
        probs = probs[:, self.components]

        opt_sel = np.zeros(len(data_col), dtype='int')
        for i in range(len(data_col)):
            pp = probs[i] + 1e-6
            pp = pp / sum(pp)
            opt_sel[i] = np.random.choice(np.arange(n_opts), p=pp)

        idx = np.arange((len(features)))
        features = features[idx, opt_sel].reshape([-1, 1])
        features = np.clip(features, -.99, .99) 
        probs_onehot = np.zeros_like(probs)
        probs_onehot[np.arange(len(probs)), opt_sel] = 1

        re_ordered_phot = np.zeros_like(probs_onehot)
        
        col_sums = probs_onehot.sum(axis=0)
        

        n = probs_onehot.shape[1]
        largest_indices = np.argsort(-1*col_sums)[:n]
        self.ordering = largest_indices
        for id,val in enumerate(largest_indices):
            re_ordered_phot[:,id] = probs_onehot[:,val]
    
        
        return np.concatenate([features, re_ordered_phot], axis=1)
        

    def inverse_transform(self, data,st):
        u = data[:, st]
        v = data[:, st + 1:st + 1 + np.sum(self.components)]
      
        v_re_ordered = np.zeros_like(v)

        for id,val in enumerate(self.ordering):
            v_re_ordered[:,val] = v[:,id]
        
        v = v_re_ordered

        u = np.clip(u, -1, 1)
        v_t = np.ones((data.shape[0], self.n_clusters)) * -100
        v_t[:, self.components] = v
        v = v_t
        
        means = self.model.means_.reshape([-1])
        stds = np.sqrt(self.model.covariances_).reshape([-1])
        p_argmax = np.argmax(v, axis=1)
        std_t = stds[p_argmax]
        mean_t = means[p_argmax]
        tmp = u * 4 * std_t + mean_t
        
        new_st = st + 1 + np.sum(self.components)
        return tmp, new_st, []

    def inverse_transform_static(self, data, transformer,st,device,n_clusters=10):
        components = transformer.get_component()
        order = transformer.get_ordering()
        model = transformer.get_model()

        components = torch.tensor(components, dtype=torch.bool).to(device)
        u = data[:, st]  # Then tahn component for "concatenating" the different components
        v = data[:, st + 1:st + 1 + torch.sum(components).item()]  # The different components
        

        # Ensure order is a tensor
        order = torch.tensor(order, device=device)
        v = v[:, torch.argsort(order)]
        
        v_t = torch.ones((data.shape[0], n_clusters), device=device) * -float('inf')
        v_t[:, components] = v
        v = v_t

        p_argmax = torch.argmax(v, dim=1)
        means = torch.tensor(model.means_.reshape([-1]), device=device)
        stds = torch.tensor(model.covariances_.reshape([-1]), device=device)
        stds = torch.sqrt(stds)    
        std_t = stds[p_argmax]
        mean_t = means[p_argmax]

        u = torch.clamp(u, -1, 1)
        tmp = u * 4 * std_t + mean_t

    
        new_st = st + 1 + torch.sum(components).item()  # Increment for next iteration
        return tmp, new_st, []




        
        
        




from abc import ABC, abstractmethod

class Column_transformer(ABC):

    def __init__(self):
        self.model = None
        self.components = None
        self.output_info = None
        self.output_dim = 0
        self.ordering = None

    def get_model(self):
        return self.model

    def get_component(self):
        return self.components

    def get_output_info(self):
        return self.output_info

    def get_output_dim(self):
        return self.output_dim

    def get_ordering(self):
        return self.ordering
        
    @abstractmethod
    def fit(self, data_col):
        raise NotImplementedError("Subclasses must implement this method")

    @abstractmethod
    def transform(self, data_col,st):
        raise NotImplementedError("Subclasses must implement this method")

    @abstractmethod
    def inverse_transform(self, data):
        raise NotImplementedError("Subclasses must implement this method")
  


import numpy as np
import torch
from scipy.stats import norm
from scipy.optimize import minimize
from .column_transformer import Column_transformer


class Gaussian_truncated_transformer(Column_transformer):
    def __init__(self, column):
        super().__init__()
        self.min = column.min()
        self.max = column.max()

        self.fit(column.to_numpy())

    def fit(self, data_col):
        self.model = None 
        self.components = None
        self.output_info = [(1, 'tanh','yes_g')]
        self.output_dim = 1

    def transform(self, data_col):
        
        self.mu, self.sigma = fit_truncated_normal(data_col, self.min, self.max)
        transformed = (data_col - self.mu) / self.sigma
        self.transformed_min = transformed.min()*1.02
        self.transformed_max = transformed.max()*1.02
        transformed_min_max = (transformed - self.transformed_min) / (self.transformed_max -self.transformed_min) * 2 - 1
        return transformed_min_max.reshape([-1, 1])
        

    def inverse_transform(self, data,st):
        transformed_min_max = data[:, st]
        transformed_min_max = np.clip(transformed_min_max, -1, 1)
        transformed = (transformed_min_max + 1) / 2 * (self.transformed_max - self.transformed_min) + self.transformed_min
        u = transformed * self.sigma + self.mu
        new_st = st + 1
        return u, new_st, []


    def inverse_transform_static(self, data, transformer, st, device, n_clusters=10): #TODO: remove n cluster when finding out how to handle it
        u = data[:, st]
        u = (u + 1) / 2
        u = torch.clamp(u, 0, 1)
        u = u * (transformer.max - transformer.min) + transformer.min
        new_st = st + 1
        return u, new_st, []



def fit_truncated_normal(data, lower_cutoff = None, upper_cutoff = None):
    if lower_cutoff is None:
        lower_cutoff = data.min()
    if upper_cutoff is None:
        upper_cutoff = data.max()
    
    truncated_data = data[(data >= lower_cutoff) & (data <= upper_cutoff)]
    init_guess = [np.mean(truncated_data), np.std(truncated_data)]
    result = minimize(
        neg_log_likelihood_two_sided,
        x0=init_guess,
        args=(truncated_data, lower_cutoff, upper_cutoff),
        bounds=[(None, None), (1e-5, None)],
        method='L-BFGS-B'
    )
    mu, sigma = result.x
    return mu, sigma


def neg_log_likelihood_two_sided(params, data, a, b):
    mu, sigma = params
    if sigma <= 0:
        return np.inf
    z = (data - mu) / sigma
    alpha = (a - mu) / sigma
    beta = (b - mu) / sigma
    logpdf = norm.logpdf(z)
    trunc_logprob = np.log(norm.cdf(beta) - norm.cdf(alpha))
    ll = np.sum(logpdf) - len(data) * np.log(sigma) - len(data) * trunc_logprob
    return -ll
        

        
        




import numpy as np
import pandas as pd
import torch
from .column_transformer import Column_transformer

class Categorical_transformer(Column_transformer):
    def __init__(self, column):
        super().__init__()
        mapper = column.value_counts().index.tolist()
        self.size = len(mapper)
        self.i2s = mapper

        self.fit(column.to_numpy())

    def fit(self, data_col):
        self.model = None
        self.components = None
        self.output_info = [(self.size, 'softmax', None)]
        self.output_dim = self.size
        unique_values, counts = np.unique(data_col, return_counts=True)
        sorted_values = unique_values[np.argsort(-counts)]
        self.i2s = sorted_values


        

    def transform(self, data_col):
        data_col = np.asarray(data_col,int)
        self.ordering = None
        value_to_index = {value: idx for idx, value in enumerate(self.i2s)}
        idx = [value_to_index[val] for val in data_col]

        col_t = np.zeros([len(data_col), self.size])
        col_t[np.arange(len(data_col)), idx] = 1
        return col_t

    def inverse_transform(self, data_col,st):
        # data_col = np.asarray(data_col,int)
        u = data_col[:, st:st + self.size]
        idx = np.argmax(u, axis=1)
        labels = self.i2s[idx]
        new_st = st + self.size
        return idx, new_st, []



    
    def inverse_transform_static(self, data, transformer, st, device,n_clusters=10): #TODO: remove n cluster when finding out how to handle this
        
        idx = torch.argmax(data[:, st:st + transformer.size], dim=1)
        new_st = st + transformer.size
        return idx, new_st, []


import numpy as np
import pandas as pd
import torch
from sklearn.mixture import BayesianGaussianMixture
from .column_transformer import Column_transformer

class Mixed_data_transformer(Column_transformer):

    def __init__(self,column,modals,n_clusters=10,eps=0.005):
        super().__init__()
        self.modals = modals
        self.n_clusters = n_clusters
        self.eps = eps
        self.min = column.min()
        self.max = column.max()
        self.nan_replacement = -9999999

        self.fit(column.to_numpy())

    def fit(self, data_col):

        gm1 = BayesianGaussianMixture(
            n_components = self.n_clusters, 
            weight_concentration_prior_type='dirichlet_process',
            weight_concentration_prior=0.001, max_iter=100,
            n_init=1,random_state=42)
        gm2 = BayesianGaussianMixture(
            n_components = self.n_clusters,
            weight_concentration_prior_type='dirichlet_process',
            weight_concentration_prior=0.001, max_iter=100,
            n_init=1,random_state=42)

        if np.nan in self.modals: self.modals[self.modals.index(np.nan)] = self.nan_replacement
        data_col = np.where(np.isnan(data_col), self.nan_replacement, data_col)
        gm1.fit(data_col.reshape([-1, 1]))

        self.filter_arr = ~np.isin(data_col, self.modals) # Find the indices of elements in data[:, id_] that are not in modal (aka that are not in the normal continuous data)
        data_continuous = self.get_continuous_data(data_col, self.filter_arr) # The observations where we have continuous data, excluding the modal/categorical data
        gm2.fit(data_continuous)

        component_assignments = gm2.predict(data_continuous) # Assigns each observation to a component

        mode_freq = pd.Series(component_assignments).value_counts().keys() # Calculate the frequency of each component assignment

        # self.filter_arr.append(filter_arr)
        self.model = gm2
        self.model0 = gm1 #TODO: remove thsi is tamp
       
        old_comp = gm2.weights_ > self.eps # To prevent the model from overfitting? removes components that are less than eps
          
        # The final components are the ones that are frequent and have a weigh greater than the eps cutoff
        self.components = [(i in mode_freq) & old_comp[i] for i in range(self.n_clusters)]


        self.output_info = [(1, 'tanh',"no_g"), (np.sum(self.components) + len(self.modals), 'softmax', None)]
        self.output_dim = 1 + np.sum(self.components) + len(self.modals)

    def transform(self, data_col):
        data_size = len(data_col)
        

        data_col = np.where(np.isnan(data_col), self.nan_replacement, data_col)
        filter_arr = self.filter_arr
        data_filter_col = self.get_continuous_data(data_col, self.filter_arr)
        
        

        #####################################

        means_0 = self.model0.means_.reshape([-1])
        stds_0 = np.sqrt(self.model0.covariances_).reshape([-1])

        zero_std_list = []
        means_needed = []
        stds_needed = []

        for mode in self.modals:
            if mode!=-9999999:
                dist = []
                for idx,val in enumerate(list(means_0.flatten())):
                    dist.append(abs(mode-val))
                index_min = np.argmin(np.array(dist))
                zero_std_list.append(index_min)
            else: continue

        for idx in zero_std_list:
            means_needed.append(means_0[idx])
            stds_needed.append(stds_0[idx])
        
        
        mode_vals = []

        for i,j,k in zip(self.modals,means_needed,stds_needed):
            this_val  = np.abs(i - j) / (4*k)
            mode_vals.append(this_val)
        
        if -9999999 in self.modals:
            mode_vals.append(0)

        #############################

        means = self.model.means_.reshape((1, self.n_clusters))
        stds = np.sqrt(self.model.covariances_).reshape((1, self.n_clusters))
        features = np.empty(shape=(len(data_filter_col),self.n_clusters))
        
        features = (data_filter_col - means) / (4 * stds)

        probs = self.model.predict_proba(data_filter_col.reshape([-1, 1]))

        n_opts = sum(self.components) 
        features = features[:, self.components]
        probs = probs[:, self.components]
        
        opt_sel = np.zeros(len(data_filter_col), dtype='int')
        for i in range(len(data_filter_col)):
            pp = probs[i] + 1e-6
            pp = pp / sum(pp)
            opt_sel[i] = np.random.choice(np.arange(n_opts), p=pp)
        idx = np.arange((len(features)))
        features = features[idx, opt_sel].reshape([-1, 1])
        features = np.clip(features, -.99, .99)
        probs_onehot = np.zeros_like(probs)
        probs_onehot[np.arange(len(probs)), opt_sel] = 1
        extra_bits = np.zeros([len(data_filter_col), len(self.modals)])
        temp_probs_onehot = np.concatenate([extra_bits,probs_onehot], axis = 1)
        final = np.zeros([len(data_col), 1 + probs_onehot.shape[1] + len(self.modals)])
        features_curser = 0
        for idx, val in enumerate(data_col): #TODO: This should be as data_col instead, since now we are using the original data, this might fix the pr
            if val in self.modals:
                category_ = list(map(self.modals.index, [val]))[0]
                final[idx, 0] = mode_vals[category_]
                final[idx, (category_+1)] = 1
            
            else:
                final[idx, 0] = features[features_curser]
                final[idx, (1+len(self.modals)):] = temp_probs_onehot[features_curser][len(self.modals):]
                features_curser = features_curser + 1
        
        just_onehot = final[:,1:]
        re_ordered_jhot= np.zeros_like(just_onehot)
        n = just_onehot.shape[1]
        col_sums = just_onehot.sum(axis=0)
        largest_indices = np.argsort(-1*col_sums)[:n]
        self.ordering = largest_indices
        for id,val in enumerate(largest_indices):
            re_ordered_jhot[:,id] = just_onehot[:,val]
        final_features = final[:,0].reshape([-1, 1])
        return np.concatenate([final_features, re_ordered_jhot], axis=1)
        

    def inverse_transform(self, data,st):
        u = data[:, st]
        full_v = data[:,(st+1):(st+1)+len(self.modals)+np.sum(self.components)]
        
        full_v_re_ordered = np.zeros_like(full_v)

        for id,val in enumerate(self.ordering):
            full_v_re_ordered[:,val] = full_v[:,id]
            
        full_v = full_v_re_ordered

        
        mixed_v = full_v[:,:len(self.modals)]
        v = full_v[:,-np.sum(self.components):]

        u = np.clip(u, -1, 1)
        v_t = np.ones((data.shape[0], self.n_clusters)) * -100
        v_t[:, self.components] = v
        v = np.concatenate([mixed_v,v_t], axis=1)

        

        means = self.model.means_.reshape([-1]) 
        stds = np.sqrt(self.model.covariances_).reshape([-1]) 
        p_argmax = np.argmax(v, axis=1)

        result = np.zeros_like(u)

        for idx in range(len(data)):
            if p_argmax[idx] < len(self.modals):
                argmax_value = p_argmax[idx]
                result[idx] = float(list(map(self.modals.__getitem__, [argmax_value]))[0])
            else:
                std_t = stds[(p_argmax[idx]-len(self.modals))]
                mean_t = means[(p_argmax[idx]-len(self.modals))]
                result[idx] = u[idx] * 4 * std_t + mean_t
        
        result[result == self.nan_replacement] = np.nan
        invalid_ids = np.where((result < self.min) | (result > self.max))[0].tolist()
        
        new_st = 1 + np.sum(self.components) + len(self.modals)
        return result, new_st, invalid_ids


    def get_continuous_data(self,data_col,filter_arr):
        data_col = np.where(np.isnan(data_col), self.nan_replacement, data_col)
        return data_col[filter_arr].reshape([-1, 1])


    def inverse_transform_static(self, data, transformer, st, device, n_clusters=10):
        components = transformer.get_component()
        order = transformer.get_ordering()
        model = transformer.get_model()
        modals = transformer.modals

        components = torch.tensor(components, dtype=torch.bool).to(device)
        u = data[:, st]
        full_v = data[:, (st + 1):(st + 1) + len(modals) + torch.sum(components).item()]

        # Ensure order is a tensor
        order = torch.tensor(order, device=device)
        full_v = full_v[:, torch.argsort(order)]

        mixed_v = full_v[:, :len(modals)]
        v = full_v[:, -torch.sum(components).item():]

        u = torch.clamp(u, -1, 1)
        v_t = torch.ones((data.shape[0], n_clusters), device=device) * -float('inf')
        v_t[:, components] = v
        v = torch.cat([mixed_v, v_t], dim=1)

        means = torch.tensor(model[1].means_.reshape([-1]), device=device)
        stds = torch.sqrt(torch.tensor(model[1].covariances_.reshape([-1]), device=device))
        p_argmax = torch.argmax(v, dim=1)

        result = torch.zeros_like(u)
        for idx in range(len(data)):
            if p_argmax[idx] < len(modals):
                argmax_value = p_argmax[idx]
                result[idx] = float(list(map(modals.__getitem__, [argmax_value]))[0])
            else:
                std_t = stds[(p_argmax[idx] - len(modals))]
                mean_t = means[(p_argmax[idx] - len(modals))]
                result[idx] = u[idx] * 4 * std_t + mean_t

        invalid_ids = torch.where((result < transformer.min) | (result > transformer.max))[0].tolist()
        new_st = 1 + torch.sum(components).item() + len(modals)
        return result, new_st, invalid_ids

        
        

        


import numpy as np
import pandas as pd
import torch
from sklearn.mixture import BayesianGaussianMixture

class DataTransformer():
    
    def __init__(self, train_data=pd.DataFrame, categorical_list=[], mixed_dict={}, general_list=[], non_categorical_list=[], n_clusters=10, eps=0.005):
        self.meta = None
        self.n_clusters = n_clusters
        self.eps = eps
        self.train_data = train_data
        self.categorical_columns= categorical_list
        self.mixed_columns= mixed_dict
        self.general_columns = general_list
        self.non_categorical_columns= non_categorical_list
        
    def get_metadata(self):
        
        meta = []
    
        for index in range(self.train_data.shape[1]):
            column = self.train_data.iloc[:,index]
            if index in self.categorical_columns:
                if index in self.non_categorical_columns:
                    meta.append({
                      "name": index,
                      "type": "continuous",
                      "min": column.min(),
                      "max": column.max(),
                    })
                else:
                    mapper = column.value_counts().index.tolist()
                    meta.append({
                        "name": index,
                        "type": "categorical",
                        "size": len(mapper),
                        "i2s": mapper
                    })

            elif index in self.mixed_columns.keys():
                meta.append({
                    "name": index,
                    "type": "mixed",
                    "min": column.min(),
                    "max": column.max(),
                    "modal": self.mixed_columns[index]
                })
            else:
                meta.append({
                    "name": index,
                    "type": "continuous",
                    "min": column.min(),
                    "max": column.max(),
                })            

        return meta

    def get_metadata2(self):
        
        meta = []
    
        for col in self.train_data:
            column = self.train_data[col]
            if col in self.categorical_columns:
                
                mapper = column.value_counts().index.tolist()
                meta.append({
                    "name": col,
                    "type": "categorical",
                    "size": len(mapper),
                    "i2s": mapper
                })

            elif col in self.mixed_columns.keys():
                meta.append({
                    "name": col,
                    "type": "mixed",
                    #"min": column.min(),
                    #"max": column.max(),
                    "modal": self.mixed_columns[col]
                })
            else:
                meta.append({
                    "name": col,
                    "type": "continuous",
                    "min": column.min(),
                    "max": column.max(),
                })            

        return meta

    def fit(self):
        data = self.train_data.values
        self.meta = self.get_metadata2()
        model = []
        self.ordering = []
        self.output_info = []
        self.output_dim = 0
        self.components = []
        self.filter_arr = []
        for id_, info in enumerate(self.meta):
            if info['type'] == "continuous":
                if info["name"] not in self.general_columns:
                  gm = BayesianGaussianMixture(
                      n_components = self.n_clusters, 
                      weight_concentration_prior_type='dirichlet_process',
                      weight_concentration_prior=0.001, 
                      max_iter=100,n_init=1, random_state=42)
                  gm.fit(data[:, id_].reshape([-1, 1]))
                  mode_freq = (pd.Series(gm.predict(data[:, id_].reshape([-1, 1]))).value_counts().keys())
                  model.append(gm)
                  old_comp = gm.weights_ > self.eps
                  comp = []
                  for i in range(self.n_clusters):
                      if (i in (mode_freq)) & old_comp[i]:
                          comp.append(True)
                      else:
                          comp.append(False)
                  self.components.append(comp) 
                  self.output_info += [(1, 'tanh','no_g'), (np.sum(comp), 'softmax')]
                  self.output_dim += 1 + np.sum(comp)
                else:
                  model.append(None)
                  self.components.append(None)
                  self.output_info += [(1, 'tanh','yes_g')]
                  self.output_dim += 1
            
            elif info['type'] == "mixed":
                """
                gm1 = BayesianGaussianMixture(
                    n_components = self.n_clusters, 
                    weight_concentration_prior_type='dirichlet_process',
                    weight_concentration_prior=0.001, max_iter=100,
                    n_init=1,random_state=42)
                """
                gm2 = BayesianGaussianMixture(
                    n_components = self.n_clusters,
                    weight_concentration_prior_type='dirichlet_process',
                    weight_concentration_prior=0.001, max_iter=100,
                    n_init=1,random_state=42)
                
                #gm1.fit(data[:, id_].reshape([-1, 1]))
                
                filter_arr = []
                for element in data[:, id_]:
                    if element not in info['modal']:
                        filter_arr.append(True)
                    else:
                        filter_arr.append(False)
                fitlered_data = data[:, id_][filter_arr]
                info["min"] = fitlered_data.min()
                info["max"] = fitlered_data.max()
                gm2.fit(fitlered_data.reshape([-1, 1]))
                mode_freq = (pd.Series(gm2.predict(fitlered_data.reshape([-1, 1]))).value_counts().keys())
                self.filter_arr.append(filter_arr)
                model.append((None,gm2))
               
                old_comp = gm2.weights_ > self.eps
                  
                comp = []
                  
                for i in range(self.n_clusters):
                    if (i in (mode_freq)) & old_comp[i]:
                        comp.append(True)
                    else:
                        comp.append(False)

                self.components.append(comp)

                self.output_info += [(1, 'tanh',"no_g"), (np.sum(comp) + len(info['modal']), 'softmax')]
                self.output_dim += 1 + np.sum(comp) + len(info['modal'])
            else:
                model.append(None)
                self.components.append(None)
                self.output_info += [(info['size'], 'softmax')]
                self.output_dim += info['size']
        self.model = model

    def transform(self, data, ispositive = False, positive_list = None):
        values = []
        mixed_counter = 0
        for id_, info in enumerate(self.meta):
            current = data[:, id_]
            if info['type'] == "continuous":
                if info["name"] not in self.general_columns:
                    
                    
                    current = current.reshape([-1, 1])
                    means = self.model[id_].means_.reshape((1, self.n_clusters))
                    stds = np.sqrt(self.model[id_].covariances_).reshape((1, self.n_clusters))
                    features = np.empty(shape=(len(current),self.n_clusters))
                    if ispositive == True:
                        if id_ in positive_list:
                            features = np.abs(current - means) / (4 * stds)
                    else:
                        features = (current - means) / (4 * stds)

                    probs = self.model[id_].predict_proba(current.reshape([-1, 1]))
                    n_opts = sum(self.components[id_])
                    features = features[:, self.components[id_]]
                    probs = probs[:, self.components[id_]]

                    opt_sel = np.zeros(len(data), dtype='int')
                    for i in range(len(data)):
                        pp = probs[i] + 1e-6
                        pp = pp / sum(pp)
                        opt_sel[i] = np.random.choice(np.arange(n_opts), p=pp)

                    idx = np.arange((len(features)))
                    features = features[idx, opt_sel].reshape([-1, 1])
                    features = np.clip(features, -.99, .99) 
                    probs_onehot = np.zeros_like(probs)
                    probs_onehot[np.arange(len(probs)), opt_sel] = 1

                    re_ordered_phot = np.zeros_like(probs_onehot)
                    
                    col_sums = probs_onehot.sum(axis=0)
                    

                    n = probs_onehot.shape[1]
                    largest_indices = np.argsort(-1*col_sums)[:n]
                    self.ordering.append(largest_indices)
                    for id,val in enumerate(largest_indices):
                        re_ordered_phot[:,id] = probs_onehot[:,val]
                    
                    
                    values += [features, re_ordered_phot]

                    
                  
                else:
                  
                  self.ordering.append(None)
                  
                  if id_ in self.non_categorical_columns:
                    info['min'] = -1e-3
                    info['max'] = info['max'] + 1e-3
                    
                  current = (current - (info['min'])) / (info['max'] - info['min'])
                  current = current * 2 - 1 
                  current = current.reshape([-1, 1])
                  values.append(current)

            elif info['type'] == "mixed":
                """
                means_0 = self.model[id_][0].means_.reshape([-1])
                stds_0 = np.sqrt(self.model[id_][0].covariances_).reshape([-1])

                zero_std_list = []
                means_needed = []
                stds_needed = []

                for mode in info['modal']:
                    if mode!=-9999999:
                        dist = []
                        for idx,val in enumerate(list(means_0.flatten())):
                            dist.append(abs(mode-val))
                        index_min = np.argmin(np.array(dist))
                        zero_std_list.append(index_min)
                    else: continue

                for idx in zero_std_list:
                    means_needed.append(means_0[idx])
                    stds_needed.append(stds_0[idx])
                
                
                mode_vals = []

                for i,j,k in zip(info['modal'],means_needed,stds_needed):
                    this_val  = np.abs(i - j) / (4*k)
                    mode_vals.append(this_val)
                
                if -9999999 in info["modal"]:
                    mode_vals.append(0)
                """
                current = current.reshape([-1, 1])
                filter_arr = self.filter_arr[mixed_counter]
                current = current[filter_arr]
                
                means = self.model[id_][1].means_.reshape((1, self.n_clusters))
                stds = np.sqrt(self.model[id_][1].covariances_).reshape((1, self.n_clusters))
                features = np.empty(shape=(len(current),self.n_clusters))
                if ispositive == True:
                    if id_ in positive_list:
                        features = np.abs(current - means) / (4 * stds)
                else:
                    features = (current - means) / (4 * stds)

                probs = self.model[id_][1].predict_proba(current.reshape([-1, 1]))

                n_opts = sum(self.components[id_]) # 8
                features = features[:, self.components[id_]]
                probs = probs[:, self.components[id_]]
                
                opt_sel = np.zeros(len(current), dtype='int')
                for i in range(len(current)):
                    pp = probs[i] + 1e-6
                    pp = pp / sum(pp)
                    opt_sel[i] = np.random.choice(np.arange(n_opts), p=pp)
                idx = np.arange((len(features)))
                features = features[idx, opt_sel].reshape([-1, 1])
                features = np.clip(features, -.99, .99)
                probs_onehot = np.zeros_like(probs)
                probs_onehot[np.arange(len(probs)), opt_sel] = 1
                extra_bits = np.zeros([len(current), len(info['modal'])])
                temp_probs_onehot = np.concatenate([extra_bits,probs_onehot], axis = 1)
                final = np.zeros([len(data), 1 + probs_onehot.shape[1] + len(info['modal'])])
                features_curser = 0
                for idx, val in enumerate(data[:, id_]):
                    if val in info['modal']:
                        category_ = list(map(info['modal'].index, [val]))[0]
                        final[idx, 0] = 0 #mode_vals[category_]
                        final[idx, (category_+1)] = 1
                    
                    else:
                        final[idx, 0] = features[features_curser]
                        final[idx, (1+len(info['modal'])):] = temp_probs_onehot[features_curser][len(info['modal']):]
                        features_curser = features_curser + 1
               
                just_onehot = final[:,1:]
                re_ordered_jhot= np.zeros_like(just_onehot)
                n = just_onehot.shape[1]
                col_sums = just_onehot.sum(axis=0)
                largest_indices = np.argsort(-1*col_sums)[:n]
                self.ordering.append(largest_indices)
                for id,val in enumerate(largest_indices):
                      re_ordered_jhot[:,id] = just_onehot[:,val]
                final_features = final[:,0].reshape([-1, 1])
                values += [final_features, re_ordered_jhot]
                mixed_counter = mixed_counter + 1
    
            else:
                self.ordering.append(None)
                col_t = np.zeros([len(data), info['size']])
                idx = list(map(info['i2s'].index, current))
                col_t[np.arange(len(data)), idx] = 1
                values.append(col_t)
                
        return np.concatenate(values, axis=1)

    def inverse_transform(self, data):
        data_t = np.zeros([len(data), len(self.meta)], dtype=object)
        invalid_ids = []
        st = 0
        for id_, info in enumerate(self.meta):
            if info['type'] == "continuous":
                if info["name"] not in self.general_columns:
                
                    
                    u = data[:, st]
                    v = data[:, st + 1:st + 1 + np.sum(self.components[id_])]
                    order = self.ordering[id_] 
                    v_re_ordered = np.zeros_like(v)

                    for id,val in enumerate(order):
                        v_re_ordered[:,val] = v[:,id]
                    
                    v = v_re_ordered

                    u = np.clip(u, -1, 1)
                    v_t = np.ones((data.shape[0], self.n_clusters)) * -100
                    v_t[:, self.components[id_]] = v
                    v = v_t
                    st += 1 + np.sum(self.components[id_])
                    means = self.model[id_].means_.reshape([-1])
                    stds = np.sqrt(self.model[id_].covariances_).reshape([-1])
                    p_argmax = np.argmax(v, axis=1)
                    std_t = stds[p_argmax]
                    mean_t = means[p_argmax]
                    tmp = u * 4 * std_t + mean_t
                                    
                    for idx,val in enumerate(tmp):
                        if (val < info["min"]) | (val > info['max']):
                            invalid_ids.append(idx)
                    
                
                    data_t[:, id_] = tmp
                   

                else:
                  u = data[:, st]
                  u = (u + 1) / 2
                  u = np.clip(u, 0, 1)
                  u = u * (info['max'] - info['min']) + info['min']
                  if id_ in self.non_categorical_columns:
                    data_t[:, id_] = np.round(u)
                  else: data_t[:, id_] = u

                  st += 1

            elif info['type'] == "mixed":

                u = data[:, st]
                full_v = data[:,(st+1):(st+1)+len(info['modal'])+np.sum(self.components[id_])]
                order = self.ordering[id_]
                full_v_re_ordered = np.zeros_like(full_v)

                for id,val in enumerate(order):
                    full_v_re_ordered[:,val] = full_v[:,id]
                  
                full_v = full_v_re_ordered

                
                mixed_v = full_v[:,:len(info['modal'])]
                v = full_v[:,-np.sum(self.components[id_]):]

                u = np.clip(u, -1, 1)
                v_t = np.ones((data.shape[0], self.n_clusters)) * -100
                v_t[:, self.components[id_]] = v
                v = np.concatenate([mixed_v,v_t], axis=1)

                st += 1 + np.sum(self.components[id_]) + len(info['modal'])
                means = self.model[id_][1].means_.reshape([-1]) 
                stds = np.sqrt(self.model[id_][1].covariances_).reshape([-1]) 
                p_argmax = np.argmax(v, axis=1)

                result = np.zeros_like(u,dtype=object)
                continous_values = np.full_like(u, np.nan, dtype=float)
                for idx in range(len(data)):
                    if p_argmax[idx] < len(info['modal']):
                        argmax_value = p_argmax[idx]
                        choosen = list(map(info['modal'].__getitem__, [argmax_value]))[0]
                        result[idx] = choosen
                    else:
                        std_t = stds[(p_argmax[idx]-len(info['modal']))]
                        mean_t = means[(p_argmax[idx]-len(info['modal']))]
                        result[idx] = u[idx] * 4 * std_t + mean_t
                        continous_values[idx] = result[idx]
                        
                for idx,val in enumerate(np.full_like(u, np.nan, dtype=float)):
                     if (val < info["min"]) | (val > info['max']):
                         invalid_ids.append(idx)

                data_t[:, id_] = result

            else:
                current = data[:, st:st + info['size']]
                st += info['size']
                idx = np.argmax(current, axis=1)
                data_t[:, id_] = list(map(info['i2s'].__getitem__, idx))
            
            
        invalid_ids = np.unique(np.array(invalid_ids)) 
        all_ids = np.arange(0,len(data))
        valid_ids = list(set(all_ids) - set(invalid_ids))
            
        return data_t[valid_ids],len(invalid_ids)


class ImageTransformer():

    def __init__(self, side):
    
        self.height = side
            
    def transform(self, data):

        if self.height * self.height > len(data[0]):
            
            padding = torch.zeros((len(data), self.height * self.height - len(data[0]))).to(data.device)
            data = torch.cat([data, padding], axis=1)

        return data.view(-1, 1, self.height, self.height)

    def inverse_transform(self, data):
        
        data = data.view(-1, self.height * self.height)

        return data





from setuptools import setup, find_packages

def read_requirements():
    with open("ctab_xtra_dp/requirements.txt", encoding="utf-8") as f:
        return [line.strip() for line in f if line.strip() and not line.startswith("#")]


setup(
    name="ctab_xtra_dp",
    version="3.0.8",
    packages=find_packages(),
    include_package_data=True,
    install_requires=read_requirements(),  # List dependencies here, e.g., ["numpy", "requests"]
    description="A sample Python package",
    long_description=open("ctab_xtra_dp/README.md").read(),
    long_description_content_type="text/markdown",
    author="kem0sabe",
    author_email="martivl@stud.ntnu.no",
    url="https://github.com/Kem0sabe/CTAB_XTRA_DP_REVISED/",  # Update with your GitHub or website
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)



import numpy as np
import pandas as pd
from sklearn.mixture import BayesianGaussianMixture





class GMM_transformer():
    def __init__(self):
        pass
        

    def fit(self,data_col,n_clusters,eps):
        self.n_clusters = n_clusters
        self.eps = eps
        self.gm = BayesianGaussianMixture(
            n_components = self.n_clusters, 
            weight_concentration_prior_type='dirichlet_process',
            weight_concentration_prior=0.001, 
            max_iter=100,n_init=1, random_state=42)
        self.gm.fit(data_col.reshape([-1, 1]))
        mode_freq = (pd.Series(self.gm.predict(data_col.reshape([-1, 1]))).value_counts().keys())
        
        old_comp = self.gm.weights_ > eps
        comp = []
        for i in range(self.n_clusters):
            if (i in (mode_freq)) & old_comp[i]:
                comp.append(True)
            else:
                comp.append(False)
        
        output_info = [(1, 'tanh','no_g'), (np.sum(comp), 'softmax')]
        output_dim = 1 + np.sum(comp)
        self.comp = comp
        return self.gm, comp, output_info, output_dim

    def transform(self,data_col):
        self.min = data_col.min()
        self.max = data_col.max()

        data_col = data_col.reshape([-1, 1])
        means = self.gm.means_.reshape((1, self.n_clusters))
        stds = np.sqrt(self.gm.covariances_).reshape((1, self.n_clusters))
        features = np.empty(shape=(len(data_col),self.n_clusters))
        
        features = (data_col - means) / (4 * stds)

        probs = self.gm.predict_proba(data_col.reshape([-1, 1]))
        n_opts = sum(self.comp)
        features = features[:, self.comp]
        probs = probs[:, self.comp]

        opt_sel = np.zeros(len(data_col), dtype='int')
        for i in range(len(data_col)):
            pp = probs[i] + 1e-6
            pp = pp / sum(pp)
            opt_sel[i] = np.random.choice(np.arange(n_opts), p=pp)

        idx = np.arange((len(features)))
        features = features[idx, opt_sel].reshape([-1, 1])
        features = np.clip(features, -.99, .99) 
        probs_onehot = np.zeros_like(probs)
        probs_onehot[np.arange(len(probs)), opt_sel] = 1

        re_ordered_phot = np.zeros_like(probs_onehot)
        
        col_sums = probs_onehot.sum(axis=0)
        

        n = probs_onehot.shape[1]
        largest_indices = np.argsort(-1*col_sums)[:n]
        self.ordering = largest_indices
        for id,val in enumerate(largest_indices):
            re_ordered_phot[:,id] = probs_onehot[:,val]

        return features, re_ordered_phot


    def inverse_transform(self,data,st):
        

        u = data[:, st]
        v = data[:, st + 1:st + 1 + np.sum(self.comp)]
        order = self.ordering 
        v_re_ordered = np.zeros_like(v)

        for id,val in enumerate(order):
            v_re_ordered[:,val] = v[:,id]
        
        v = v_re_ordered

        u = np.clip(u, -1, 1)
        v_t = np.ones((data.shape[0], self.n_clusters)) * -100
        v_t[:, self.comp] = v
        v = v_t
        st += 1 + np.sum(self.comp)
        means = self.gm.means_.reshape([-1])
        stds = np.sqrt(self.gm.covariances_).reshape([-1])
        p_argmax = np.argmax(v, axis=1)
        std_t = stds[p_argmax]
        mean_t = means[p_argmax]
        tmp = u * 4 * std_t + mean_t

        invalid_ids = []       
        for idx,val in enumerate(tmp):
            if (val < self.min) | (val > self.max):
                invalid_ids.append(idx)
        return tmp , invalid_ids
